/** @module hareactive/behavior */
"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var frp_common_1 = require("./frp-common");
var Future_1 = require("./Future");
var F = require("./Future");
var IncompleteObserver = (function () {
    function IncompleteObserver() {
    }
    IncompleteObserver.prototype.beginPulling = function () {
        throw new Error("beginPulling not implemented");
    };
    IncompleteObserver.prototype.endPulling = function () {
        throw new Error("beginPushing not implemented");
    };
    IncompleteObserver.prototype.push = function (a) {
        throw new Error("push not implemented");
    };
    return IncompleteObserver;
}());
var OnlyPushObserver = (function () {
    function OnlyPushObserver(cb) {
        this.cb = cb;
    }
    ;
    OnlyPushObserver.prototype.beginPulling = function () { };
    OnlyPushObserver.prototype.endPulling = function () { };
    OnlyPushObserver.prototype.push = function (a) {
        this.cb(a);
    };
    return OnlyPushObserver;
}());
/**
 * A behavior is a value that changes over time. Conceptually it can
 * be though of as a function from time to a value. I.e. `type
 * Behavior<A> = (t: Time) => A`.
 */
var Behavior = (function () {
    function Behavior() {
        this.of = of;
        this.child = frp_common_1.noopObserver;
        this.nrOfListeners = 0;
    }
    Behavior.prototype.map = function (fn) {
        var newB = new MapBehavior(this, fn);
        this.addListener(newB);
        return newB;
    };
    Behavior.prototype.ap = function (f) {
        var newB = new ApBehavior(f, this);
        f.addListener(newB);
        this.addListener(newB);
        return newB;
    };
    Behavior.prototype.lift = function () {
        // TODO: Experiment with faster specialized `lift` implementation
        var f = arguments[0];
        switch (arguments.length - 1) {
            case 1:
                return arguments[1].map(f);
            case 2:
                return arguments[2].ap(arguments[1].map(function (a) { return function (b) { return f(a, b); }; }));
            case 3:
                return arguments[3].ap(arguments[2].ap(arguments[1].map(function (a) { return function (b) { return function (c) { return f(a, b, c); }; }; })));
        }
    };
    Behavior.prototype.chain = function (fn) {
        return new ChainBehavior(this, fn);
    };
    Behavior.prototype.endPulling = function () {
        this.pushing = true;
        this.child.endPulling();
    };
    Behavior.prototype.beginPulling = function () {
        this.pushing = false;
        this.child.beginPulling();
    };
    Behavior.prototype.subscribe = function (cb) {
        var listener = new OnlyPushObserver(cb);
        this.addListener(listener);
        return listener;
    };
    Behavior.prototype.addListener = function (c) {
        var nr = ++this.nrOfListeners;
        if (nr === 1) {
            this.child = c;
        }
        else if (nr === 2) {
            this.child = new frp_common_1.MultiObserver(this.child, c);
        }
        else {
            this.child.listeners.push(c);
        }
    };
    Behavior.prototype.removeListener = function (listener) {
        var nr = --this.nrOfListeners;
        if (nr === 0) {
            this.child = frp_common_1.noopObserver;
        }
        else if (nr === 1) {
            var l = this.child.listeners;
            this.child = l[l[0] === listener ? 1 : 0];
        }
        else {
            var l = this.child.listeners;
            // The indexOf here is O(n), where n is the number of listeners,
            // if using a linked list it should be possible to perform the
            // unsubscribe operation in constant time.
            var idx = l.indexOf(listener);
            if (idx !== -1) {
                if (idx !== l.length - 1) {
                    l[idx] = l[l.length - 1];
                }
                l.length--; // remove the last element of the list
            }
        }
    };
    return Behavior;
}());
exports.Behavior = Behavior;
Behavior.of = of;
function of(val) {
    return new ConstantBehavior(val);
}
exports.of = of;
/*
 * Impure function that gets the current value of a behavior. For a
 * pure variant see `sample`.
 */
function at(b) {
    return b.pushing === true ? b.last : b.pull();
}
exports.at = at;
/** @private */
var ConstantBehavior = (function (_super) {
    __extends(ConstantBehavior, _super);
    function ConstantBehavior(last) {
        _super.call(this);
        this.last = last;
        this.pushing = true;
    }
    ConstantBehavior.prototype.push = function () {
        throw new Error("Cannot push a value to a constant behavior");
    };
    ConstantBehavior.prototype.pull = function () {
        return this.last;
    };
    return ConstantBehavior;
}(Behavior));
/** @private */
var MapBehavior = (function (_super) {
    __extends(MapBehavior, _super);
    function MapBehavior(parent, fn) {
        _super.call(this);
        this.parent = parent;
        this.fn = fn;
        this.pushing = parent.pushing;
        if (this.pushing === true) {
            this.last = fn(at(parent));
        }
    }
    MapBehavior.prototype.push = function (a) {
        this.last = this.fn(a);
        this.child.push(this.last);
    };
    MapBehavior.prototype.pull = function () {
        return this.fn(at(this.parent));
    };
    return MapBehavior;
}(Behavior));
/**
 * Map a function over a behavior. This means that if at some point in
 * time the value of `b` is `bVal` then the value of the returned
 * behavior is `fn(bVal)`.
 */
function map(fn, b) {
    return b.map(fn);
}
exports.map = map;
/** @private */
var ChainOuter = (function (_super) {
    __extends(ChainOuter, _super);
    function ChainOuter(chainB) {
        _super.call(this);
        this.chainB = chainB;
    }
    ;
    ChainOuter.prototype.push = function (a) {
        this.chainB.pushOuter(a);
    };
    return ChainOuter;
}(IncompleteObserver));
/** @private */
var ChainBehavior = (function (_super) {
    __extends(ChainBehavior, _super);
    // private innerConsumer: Consumer<B>;
    function ChainBehavior(outer, fn) {
        _super.call(this);
        this.outer = outer;
        this.fn = fn;
        // Create the outer consumers
        this.outerConsumer = new ChainOuter(this);
        this.innerB = this.fn(at(this.outer));
        this.pushing = this.innerB.pushing;
        // Make the consumers listen to inner and outer behavior
        outer.addListener(this.outerConsumer);
        this.innerB.addListener(this);
        this.last = at(this.innerB);
    }
    ChainBehavior.prototype.pushOuter = function (a) {
        // The outer behavior has changed. This means that we will have to
        // call our function, which will result in a new inner behavior.
        // We therefore stop listening to the old inner behavior and begin
        // listening to the new one.
        this.innerB.removeListener(this);
        var newInner = this.innerB = this.fn(a);
        newInner.addListener(this);
        this.push(at(newInner));
    };
    ChainBehavior.prototype.push = function (b) {
        this.last = b;
        this.child.push(b);
    };
    ChainBehavior.prototype.pull = function () {
        return at(this.fn(at(this.outer)));
    };
    return ChainBehavior;
}(Behavior));
/** @private */
var FunctionBehavior = (function (_super) {
    __extends(FunctionBehavior, _super);
    function FunctionBehavior(fn) {
        _super.call(this);
        this.fn = fn;
        this.pushing = false;
    }
    FunctionBehavior.prototype.push = function (v) {
        throw new Error("Cannot push to a FunctionBehavior");
    };
    FunctionBehavior.prototype.pull = function () {
        return this.fn();
    };
    return FunctionBehavior;
}(Behavior));
/** @private */
var ApBehavior = (function (_super) {
    __extends(ApBehavior, _super);
    function ApBehavior(fn, val) {
        _super.call(this);
        this.fn = fn;
        this.val = val;
        this.pushing = fn.pushing && val.pushing;
        this.last = at(fn)(at(val));
    }
    ApBehavior.prototype.push = function () {
        var fn = at(this.fn);
        var val = at(this.val);
        this.last = fn(val);
        this.child.push(this.last);
    };
    ApBehavior.prototype.pull = function () {
        return at(this.fn)(at(this.val));
    };
    return ApBehavior;
}(Behavior));
/**
 * Apply a function valued behavior to a value behavior.
 *
 * @param fnB behavior of functions from `A` to `B`
 * @param valB A behavior of `A`
 * @returns Behavior of the function in `fnB` applied to the value in `valB`
 */
function ap(fnB, valB) {
    return valB.ap(fnB);
}
exports.ap = ap;
/** @private */
var SinkBehavior = (function (_super) {
    __extends(SinkBehavior, _super);
    function SinkBehavior(last) {
        _super.call(this);
        this.last = last;
        this.pushing = true;
    }
    SinkBehavior.prototype.push = function (v) {
        if (this.last !== v) {
            this.last = v;
            this.child.push(v);
        }
    };
    SinkBehavior.prototype.pull = function () {
        return this.last;
    };
    return SinkBehavior;
}(Behavior));
/**
 * Creates a behavior for imperative impure pushing.
 */
function sink(initialValue) {
    return new SinkBehavior(initialValue);
}
exports.sink = sink;
/**
 * A placeholder behavior is a behavior without any value. It is used
 * to do value recursion in `./framework.ts`.
 * @private
 */
var PlaceholderBehavior = (function (_super) {
    __extends(PlaceholderBehavior, _super);
    function PlaceholderBehavior() {
        _super.call(this);
        // `undefined` indicates that this behavior is neither pushing nor
        // pulling
        this.pushing = undefined;
    }
    PlaceholderBehavior.prototype.push = function (v) {
        this.last = v;
        this.child.push(v);
    };
    PlaceholderBehavior.prototype.pull = function () {
        return this.source.pull();
    };
    PlaceholderBehavior.prototype.replaceWith = function (b) {
        this.source = b;
        b.addListener(this);
        this.pushing = b.pushing;
        if (b.pushing === true) {
            this.pushing = false;
            this.push(at(b));
        }
        else {
            this.beginPulling();
        }
    };
    return PlaceholderBehavior;
}(Behavior));
exports.PlaceholderBehavior = PlaceholderBehavior;
function placeholder() {
    return new PlaceholderBehavior();
}
exports.placeholder = placeholder;
/** @private */
var WhenBehavior = (function (_super) {
    __extends(WhenBehavior, _super);
    function WhenBehavior(parent) {
        _super.call(this);
        this.parent = parent;
        this.pushing = true;
        parent.addListener(this);
        this.push(at(parent));
    }
    WhenBehavior.prototype.push = function (val) {
        if (val === true) {
            this.last = Future_1.Future.of({});
        }
        else {
            this.last = new Future_1.BehaviorFuture(this.parent);
        }
    };
    WhenBehavior.prototype.pull = function () {
        return this.last;
    };
    return WhenBehavior;
}(Behavior));
/**
 * Take a behavior `b` of booleans and return a behavior that for time
 * `t` contains a future that occurs when `b` is true after `t`.
 * @param b - A boolean valued behavior.
 */
function when(b) {
    return new WhenBehavior(b);
}
exports.when = when;
// FIXME: This can probably be made less ugly.
/** @private */
var SnapshotBehavior = (function (_super) {
    __extends(SnapshotBehavior, _super);
    function SnapshotBehavior(parent, future) {
        _super.call(this);
        this.parent = parent;
        if (future.occured === true) {
            // Future has occurred at some point in the past
            this.afterFuture = true;
            this.pushing = parent.pushing;
            parent.addListener(this);
            this.last = Future_1.Future.of(at(parent));
        }
        else {
            this.afterFuture = false;
            this.pushing = true;
            this.last = F.sink();
            future.listen(this);
        }
    }
    SnapshotBehavior.prototype.push = function (val) {
        if (this.afterFuture === false) {
            // The push is coming from the Future, it has just occurred.
            this.afterFuture = true;
            this.last.resolve(at(this.parent));
            this.parent.addListener(this);
        }
        else {
            // We are recieving an update from `parent` after `future` has
            // occurred.
            this.last = Future_1.Future.of(val);
        }
    };
    SnapshotBehavior.prototype.pull = function () {
        return this.last;
    };
    return SnapshotBehavior;
}(Behavior));
/**
 * Creates a future than on occurence samples the current value of the
 * behavior and occurs with that value. That is, the original value of
 * the future is overwritten with the behavior value at the time when
 * the future occurs.
 */
function snapshot(behavior, future) {
    return new SnapshotBehavior(behavior, future);
}
exports.snapshot = snapshot;
/** @private */
var SwitcherBehavior = (function (_super) {
    __extends(SwitcherBehavior, _super);
    function SwitcherBehavior(b, next) {
        _super.call(this);
        this.b = b;
        this.pushing = b.pushing;
        if (this.pushing === true) {
            this.last = at(b);
        }
        b.addListener(this);
        // FIXME: Using `bind` is hardly optimal for performance.
        next.subscribe(this.doSwitch.bind(this));
    }
    SwitcherBehavior.prototype.push = function (val) {
        this.last = val;
        this.child.push(val);
    };
    SwitcherBehavior.prototype.pull = function () {
        return at(this.b);
    };
    SwitcherBehavior.prototype.doSwitch = function (newB) {
        this.b.removeListener(this);
        this.b = newB;
        newB.addListener(this);
        if (newB.pushing === true) {
            if (this.pushing === false) {
                this.endPulling();
            }
            this.push(at(newB));
        }
        else if (this.pushing === true) {
            this.beginPulling();
        }
    };
    return SwitcherBehavior;
}(Behavior));
/**
 * From an initial behavior and a future of a behavior `switcher`
 * creates a new behavior that acts exactly like `initial` until
 * `next` occurs after which it acts like the behavior it contains.
 */
function switcher(init, next) {
    return new SwitcherBehavior(init, next);
}
exports.switcher = switcher;
/** @private */
var StepperBehavior = (function (_super) {
    __extends(StepperBehavior, _super);
    function StepperBehavior(initial, steps) {
        _super.call(this);
        this.steps = steps;
        this.pushing = true;
        this.last = initial;
        steps.addListener(this);
    }
    StepperBehavior.prototype.push = function (val) {
        this.last = val;
        this.child.push(val);
    };
    StepperBehavior.prototype.pull = function () {
        throw new Error("Cannot pull from StepperBehavior");
    };
    return StepperBehavior;
}(Behavior));
/**
 * Creates a Behavior whose value is the last occurrence in the stream.
 * @param initial the initial value that the behavior has
 * @param steps the stream that will change the value of the behavior
 */
function stepper(initial, steps) {
    return new StepperBehavior(initial, steps);
}
exports.stepper = stepper;
/** @private */
var ScanBehavior = (function (_super) {
    __extends(ScanBehavior, _super);
    function ScanBehavior(initial, fn, source) {
        _super.call(this);
        this.fn = fn;
        this.source = source;
        this.pushing = true;
        this.last = initial;
        source.addListener(this);
    }
    ScanBehavior.prototype.push = function (val) {
        this.last = this.fn(val, this.last);
        this.child.push(this.last);
    };
    ScanBehavior.prototype.pull = function () {
        throw new Error("Cannot pull from Scan");
    };
    return ScanBehavior;
}(Behavior));
/**
 * The returned behavior initially has the initial value, on each
 * occurence in `source` the function is applied to the current value
 * of the behaviour and the value of the occurence, the returned value
 * becomes the next value of the behavior.
 */
function scan(fn, init, source) {
    return fromFunction(function () { return new ScanBehavior(init, fn, source); });
}
exports.scan = scan;
/**
 * This takes an impure function that varies over time and returns a
 * pull-driven behavior. This is particulairly useful if the function
 * is contionusly changing, like `Date.now`.
 */
function fromFunction(fn) {
    return new FunctionBehavior(fn);
}
exports.fromFunction = fromFunction;
/**
 * Subscribe to a behavior in order to run imperative actions when the
 * value in the behavior changes.
 */
function subscribe(fn, b) {
    b.subscribe(fn);
}
exports.subscribe = subscribe;
var CbObserver = (function () {
    function CbObserver(_push, _beginPulling, _endPulling, source) {
        this._push = _push;
        this._beginPulling = _beginPulling;
        this._endPulling = _endPulling;
        this.source = source;
        source.addListener(this);
        // We explicitly checks for both `true` and `false` because
        // placeholder behavior is `undefined`
        if (source.pushing === false) {
            _beginPulling();
        }
        else if (source.pushing === true) {
            _push(source.last);
        }
    }
    CbObserver.prototype.push = function (a) {
        this._push(a);
    };
    CbObserver.prototype.beginPulling = function () {
        this._beginPulling();
    };
    CbObserver.prototype.endPulling = function () {
        this._endPulling();
    };
    return CbObserver;
}());
exports.CbObserver = CbObserver;
/**
 * Observe a behavior for the purpose of executing imperative actions
 * based on the value of the behavior.
 */
function observe(push, beginPulling, endPulling, b) {
    return new CbObserver(push, beginPulling, endPulling, b);
}
exports.observe = observe;
/**
 * Imperatively push a value into a behavior.
 */
function publish(a, b) {
    b.push(a);
}
exports.publish = publish;
function isBehavior(b) {
    return b instanceof Behavior;
}
exports.isBehavior = isBehavior;
var TimeFromBehavior = (function (_super) {
    __extends(TimeFromBehavior, _super);
    function TimeFromBehavior() {
        _super.call(this);
        this.startTime = Date.now();
        this.pushing = false;
    }
    TimeFromBehavior.prototype.push = function () {
        throw new Error("Cannot push to TimeFromBehavior");
    };
    TimeFromBehavior.prototype.pull = function () {
        return Date.now() - this.startTime;
    };
    return TimeFromBehavior;
}(Behavior));
/**
 * A behavior whose value is the number of milliseconds elapsed win
 * UNIX epoch. I.e. its current value is equal to the value got by
 * calling `Date.now`.
 */
exports.time = fromFunction(Date.now);
/**
 * A behavior giving access to continous time. When sampled the outer
 * behavior gives a behavior with values that contain the difference
 * between the current sample time and the time at which the outer
 * behavior was sampled.
 */
exports.timeFrom = fromFunction(function () { return new TimeFromBehavior(); });
function lift() {
    return arguments[1].lift.apply(undefined, arguments);
}
exports.lift = lift;
//# sourceMappingURL=Behavior.js.map