/** @module hareactive/stream */
import { MapFunction, SubscribeFunction, ScanFunction, FilterFunction, Consumer } from "./frp-common";
import { Behavior } from "./Behavior";
/**
 * A stream is a list of occurences over time. Each occurence happens
 * at a discrete point in time and has an associated value.
 * Semantically it is a list `type Stream<A> = [Time, A]`.
 */
export declare abstract class Stream<A> {
    eventListeners: any[];
    data: any;
    push: (a: any, s: Stream<any>, b: any) => void;
    constructor(data: any, push: any);
    publish(a: A): void;
    subscribe(fn: (a: A) => void): void;
    map<B>(fn: MapFunction<A, B>): Stream<B>;
    filter(fn: FilterFunction<A>): Stream<A>;
    scanS<B>(fn: ScanFunction<A, B>, startingValue: B): Behavior<Stream<B>>;
    scan<B>(fn: ScanFunction<A, B>, init: B): Behavior<Behavior<B>>;
    unlisten(listener: Consumer<any>): void;
}
/**
 * @param fn A predicate function that returns a boolean for `A`.
 * @param stream The stream to filter.
 * @returns Stream that only contains the occurences from `stream`
 * for which `fn` returns true.
 */
export declare function filter<A>(fn: (a: A) => boolean, stream: Stream<A>): Stream<A>;
/**
 * The returned  initially has the initial value, on each
 * occurence in `source` the function is applied to the current value
 * of the behaviour and the value of the occurence, the returned value
 * becomes the next value of the behavior.
 */
export declare function scanS<A, B>(fn: ScanFunction<A, B>, startingValue: B, stream: Stream<A>): Behavior<Stream<B>>;
/**
 * Takes a stream valued behavior and returns at stream that emits
 * values from the current stream at the behavior.
 */
/**
 * @returns A stream that never has any occurrences.
 */
export declare function empty<A>(): Stream<A>;
export declare function subscribe<A>(fn: SubscribeFunction<A>, stream: Stream<A>): void;
export declare function publish<A>(a: A, stream: Stream<A>): void;
export declare function map<A, B>(fn: MapFunction<A, B>, stream: Stream<A>): Stream<B>;
export declare function isStream(obj: any): boolean;
