"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var io_1 = require("jabz/io");
var monad_1 = require("jabz/monad");
var Future_1 = require("./Future");
var Behavior_1 = require("./Behavior");
var Stream_1 = require("./Stream");
/**
 * The Now monad represents a computation that takes place in a given
 * moment and where the moment will always be now when the computation
 * is run.
 */
var Now = (function () {
    function Now() {
        this.multi = false;
    }
    Now.prototype.of = function (b) {
        return new OfNow(b);
    };
    Now.of = function (b) {
        return new OfNow(b);
    };
    Now.prototype.chain = function (f) {
        return new ChainNow(this, f);
    };
    return Now;
}());
Now.multi = false;
Now = __decorate([
    monad_1.monad
], Now);
exports.Now = Now;
var OfNow = (function (_super) {
    __extends(OfNow, _super);
    function OfNow(value) {
        _super.call(this);
        this.value = value;
    }
    OfNow.prototype.run = function () {
        return this.value;
    };
    return OfNow;
}(Now));
var ChainNow = (function (_super) {
    __extends(ChainNow, _super);
    function ChainNow(first, f) {
        _super.call(this);
        this.first = first;
        this.f = f;
    }
    ChainNow.prototype.run = function () {
        return this.f(this.first.run()).run();
    };
    return ChainNow;
}(Now));
var AsyncNow = (function (_super) {
    __extends(AsyncNow, _super);
    function AsyncNow(comp) {
        _super.call(this);
        this.comp = comp;
    }
    AsyncNow.prototype.run = function () {
        return Future_1.fromPromise(io_1.runIO(this.comp));
    };
    return AsyncNow;
}(Now));
/**
 * Run an asynchronous IO action and return a future in the Now monad
 * that resolves with the eventual result of the IO action once it
 * completes. This function is what allows the Now monad to execute
 * imperative actions in a way that is pure and integrated with FRP.
 */
function async(comp) {
    return new AsyncNow(comp);
}
exports.async = async;
var SampleNow = (function (_super) {
    __extends(SampleNow, _super);
    function SampleNow(b) {
        _super.call(this);
        this.b = b;
    }
    SampleNow.prototype.run = function () {
        return Behavior_1.at(this.b);
    };
    return SampleNow;
}(Now));
/**
 * Returns the current value of a behavior in the Now monad. This is
 * possible because computations in the Now monad have an associated
 * point in time.
 */
function sample(b) {
    return new SampleNow(b);
}
exports.sample = sample;
var PerformIOStream = (function (_super) {
    __extends(PerformIOStream, _super);
    function PerformIOStream(s) {
        _super.call(this);
        s.addListener(this);
    }
    PerformIOStream.prototype.push = function (io) {
        var _this = this;
        io_1.runIO(io).then(function (a) { return _this.child.push(a); });
    };
    return PerformIOStream;
}(Stream_1.Stream));
var PerformStreamNow = (function (_super) {
    __extends(PerformStreamNow, _super);
    function PerformStreamNow(s) {
        _super.call(this);
        this.s = s;
    }
    PerformStreamNow.prototype.run = function () {
        return new PerformIOStream(this.s);
    };
    return PerformStreamNow;
}(Now));
/**
 * Takes a stream of `IO` actions and return a stream in a now
 * computation. When run the now computation executes each `IO` action
 * and delivers their result into the created stream.
 */
function performStream(s) {
    return new PerformStreamNow(s);
}
exports.performStream = performStream;
function run(now) {
    return now.run();
}
var PlanNow = (function (_super) {
    __extends(PlanNow, _super);
    function PlanNow(future) {
        _super.call(this);
        this.future = future;
    }
    PlanNow.prototype.run = function () {
        return this.future.map(run);
    };
    return PlanNow;
}(Now));
/**
 * Convert a future now computation into a now computation of a
 * future. This function is what allows a Now-computation to reach
 * beyond the current moment that it is running in.
 */
function plan(future) {
    return new PlanNow(future);
}
exports.plan = plan;
/**
 * Run the given Now-computation. The returned promise resolves once
 * the future that is the result of running the now computation
 * occurs. This is an impure function and should not be used in normal
 * application code.
 */
function runNow(now) {
    return new Promise(function (resolve, reject) {
        now.run().subscribe(resolve);
    });
}
exports.runNow = runNow;
//# sourceMappingURL=Now.js.map