"use strict";
var NoopConsumer = (function () {
    function NoopConsumer() {
    }
    NoopConsumer.prototype.push = function () { };
    ;
    return NoopConsumer;
}());
exports.NoopConsumer = NoopConsumer;
exports.noopConsumer = new NoopConsumer();
var MultiConsumer = (function () {
    function MultiConsumer(c1, c2) {
        this.listeners = [c1, c2];
    }
    MultiConsumer.prototype.push = function (a) {
        for (var i = 0; i < this.listeners.length; ++i) {
            this.listeners[i].push(a);
        }
    };
    return MultiConsumer;
}());
exports.MultiConsumer = MultiConsumer;
var Reactive = (function () {
    function Reactive() {
        this.child = exports.noopConsumer;
        this.nrOfListeners = 0;
    }
    Reactive.prototype.subscribe = function (fn) {
        var listener = { push: fn };
        this.addListener(listener);
        return listener;
    };
    Reactive.prototype.addListener = function (c) {
        var nr = ++this.nrOfListeners;
        if (nr === 1) {
            this.child = c;
        }
        else if (nr === 2) {
            this.child = new MultiConsumer(this.child, c);
        }
        else {
            this.child.listeners.push(c);
        }
    };
    Reactive.prototype.removeListener = function (listener) {
        var nr = --this.nrOfListeners;
        if (nr === 0) {
            this.child = exports.noopConsumer;
        }
        else if (nr === 1) {
            var l = this.child.listeners;
            this.child = l[l[0] === listener ? 1 : 0];
        }
        else {
            var l = this.child.listeners;
            // The indexOf here is O(n), where n is the number of listeners,
            // if using a linked list it should be possible to perform the
            // unsubscribe operation in constant time.
            var idx = l.indexOf(listener);
            if (idx !== -1) {
                if (idx !== l.length - 1) {
                    l[idx] = l[l.length - 1];
                }
                l.length--; // remove the last element of the list
            }
        }
    };
    return Reactive;
}());
exports.Reactive = Reactive;
var MultiObserver = (function () {
    function MultiObserver(c1, c2) {
        this.listeners = [c1, c2];
    }
    MultiObserver.prototype.beginPulling = function () {
        for (var i = 0; i < this.listeners.length; ++i) {
            this.listeners[i].beginPulling();
        }
    };
    MultiObserver.prototype.endPulling = function () {
        for (var i = 0; i < this.listeners.length; ++i) {
            this.listeners[i].endPulling();
        }
    };
    MultiObserver.prototype.push = function (a) {
        for (var i = 0; i < this.listeners.length; ++i) {
            this.listeners[i].push(a);
        }
    };
    return MultiObserver;
}());
exports.MultiObserver = MultiObserver;
var NoopObserver = (function () {
    function NoopObserver() {
    }
    NoopObserver.prototype.beginPulling = function () { };
    NoopObserver.prototype.endPulling = function () { };
    NoopObserver.prototype.push = function (a) { };
    return NoopObserver;
}());
exports.noopObserver = new NoopObserver();
//# sourceMappingURL=frp-common.js.map