"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var monad_1 = require("jabz/monad");
/**
 * A future is a thing that occurs at some point in time with a value.
 * It can be understood as a pair consisting of the time the future
 * occurs and its associated value. It is quite like a JavaScript
 * promise.
 */
var Future = (function () {
    function Future() {
        this.multi = false;
        this.listeners = [];
    }
    Future.prototype.listen = function (o) {
        if (this.occured !== true) {
            this.listeners.push(o);
        }
        else {
            o.push(this.value);
        }
    };
    Future.prototype.subscribe = function (f) {
        new Subscription(f, this);
    };
    Future.prototype.resolve = function (val) {
        this.occured = true;
        this.value = val;
        var listeners = this.listeners;
        for (var i = 0, l = listeners.length; i < l; ++i) {
            listeners[i].push(val);
        }
    };
    // A future is a functor, when the future occurs we can feed its
    // result through the mapping function
    Future.prototype.map = function (f) {
        return new MapFuture(f, this);
    };
    Future.prototype.mapTo = function (b) {
        return new MapToFuture(b, this);
    };
    // A future is an applicative. `of` gives a future that has always
    // occured at all points in time.
    Future.of = function (b) {
        return new PureFuture(b);
    };
    Future.prototype.of = function (b) {
        return new PureFuture(b);
    };
    Future.prototype.lift = function (f) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        return f.length === 1 ? new MapFuture(f, args[0])
            : new LiftFuture(f, args);
    };
    // A future is a monad. Once the first future occurs `chain` passes
    // its value through the chain function and the future it returns is
    // the one returned by `chain`.
    Future.prototype.chain = function (f) {
        return new ChainFuture(f, this);
    };
    return Future;
}());
Future = __decorate([
    monad_1.monad
], Future);
exports.Future = Future;
var MapFuture = (function (_super) {
    __extends(MapFuture, _super);
    function MapFuture(f, parent) {
        _super.call(this);
        this.f = f;
        this.parent = parent;
        parent.listen(this);
    }
    MapFuture.prototype.push = function (val) {
        this.resolve(this.f(val));
    };
    return MapFuture;
}(Future));
var MapToFuture = (function (_super) {
    __extends(MapToFuture, _super);
    function MapToFuture(value, parent) {
        _super.call(this);
        this.value = value;
        this.parent = parent;
        parent.listen(this);
    }
    MapToFuture.prototype.push = function (_) {
        this.resolve(this.value);
    };
    return MapToFuture;
}(Future));
var PureFuture = (function (_super) {
    __extends(PureFuture, _super);
    function PureFuture(value) {
        _super.call(this);
        this.value = value;
        this.occured = true;
    }
    PureFuture.prototype.push = function (_) {
        throw new Error("A PureFuture should never be pushed to.");
    };
    return PureFuture;
}(Future));
var LiftFuture = (function (_super) {
    __extends(LiftFuture, _super);
    function LiftFuture(f, futures) {
        _super.call(this);
        this.f = f;
        this.futures = futures;
        this.delivered = 0;
        var l = this.dependencies = futures.length;
        for (var i = 0; i < l; ++i) {
            futures[i].listen(this);
        }
    }
    LiftFuture.prototype.push = function (_) {
        var l = this.dependencies;
        if (++this.delivered === l) {
            // All the dependencies have occurred.
            for (var i = 0; i < l; ++i) {
                this.futures[i] = this.futures[i].value;
            }
            this.resolve(this.f.apply(undefined, this.futures));
        }
    };
    return LiftFuture;
}(Future));
var ChainFuture = (function (_super) {
    __extends(ChainFuture, _super);
    function ChainFuture(f, parent) {
        _super.call(this);
        this.f = f;
        this.parent = parent;
        this.parentOccurred = false;
        parent.listen(this);
    }
    ChainFuture.prototype.push = function (val) {
        if (this.parentOccurred === false) {
            // The first future occured. We can now call `f` with its value
            // and listen to the future it returns.
            this.parentOccurred = true;
            var newFuture = this.f(val);
            newFuture.listen(this);
        }
        else {
            this.resolve(val);
        }
    };
    return ChainFuture;
}(Future));
// A Sink is a producer that one can imperatively resolve.
var FutureSink = (function (_super) {
    __extends(FutureSink, _super);
    function FutureSink() {
        _super.apply(this, arguments);
    }
    FutureSink.prototype.push = function (val) {
        throw new Error("A sink should never be pushed to.");
    };
    return FutureSink;
}(Future));
function sinkFuture() {
    return new FutureSink();
}
exports.sinkFuture = sinkFuture;
// A subscription is a consumer that performs a side-effect
var Subscription = (function () {
    function Subscription(f, parent) {
        this.f = f;
        this.parent = parent;
        parent.listen(this);
    }
    Subscription.prototype.push = function (a) {
        this.f(a); // let `f` perform its side-effect.
    };
    return Subscription;
}());
function fromPromise(p) {
    var future = sinkFuture();
    p.then(future.resolve.bind(future));
    return future;
}
exports.fromPromise = fromPromise;
/**
 * Create a future from a pushing behavior. The future occurs when the
 * behavior pushes its next value. Constructing a BehaviorFuture is
 * impure and should not be done direcly.
 * @private
 */
var BehaviorFuture = (function (_super) {
    __extends(BehaviorFuture, _super);
    function BehaviorFuture(b) {
        _super.call(this);
        this.b = b;
        b.addListener(this);
    }
    BehaviorFuture.prototype.endPulling = function () {
        throw new Error("Behavior future should never switch to pushing");
    };
    BehaviorFuture.prototype.beginPulling = function () {
        throw new Error("Behavior future does not support pushing behavior");
    };
    BehaviorFuture.prototype.push = function (a) {
        this.b.removeListener(this);
        this.resolve(a);
    };
    return BehaviorFuture;
}(Future));
exports.BehaviorFuture = BehaviorFuture;
//# sourceMappingURL=future.js.map