"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var io_1 = require("jabz/io");
var monad_1 = require("jabz/monad");
var future_1 = require("./future");
var behavior_1 = require("./behavior");
var stream_1 = require("./stream");
var Now = (function () {
    function Now() {
        this.multi = false;
    }
    Now.prototype.of = function (b) {
        return new OfNow(b);
    };
    Now.of = function (b) {
        return new OfNow(b);
    };
    Now.prototype.chain = function (f) {
        return new ChainNow(this, f);
    };
    return Now;
}());
Now.multi = false;
Now = __decorate([
    monad_1.monad
], Now);
exports.Now = Now;
var OfNow = (function (_super) {
    __extends(OfNow, _super);
    function OfNow(value) {
        _super.call(this);
        this.value = value;
    }
    OfNow.prototype.run = function () {
        return this.value;
    };
    return OfNow;
}(Now));
var ChainNow = (function (_super) {
    __extends(ChainNow, _super);
    function ChainNow(first, f) {
        _super.call(this);
        this.first = first;
        this.f = f;
    }
    ChainNow.prototype.run = function () {
        return this.f(this.first.run()).run();
    };
    return ChainNow;
}(Now));
var AsyncNow = (function (_super) {
    __extends(AsyncNow, _super);
    function AsyncNow(comp) {
        _super.call(this);
        this.comp = comp;
    }
    AsyncNow.prototype.run = function () {
        return future_1.fromPromise(io_1.runIO(this.comp));
    };
    return AsyncNow;
}(Now));
function async(comp) {
    return new AsyncNow(comp);
}
exports.async = async;
var SampleNow = (function (_super) {
    __extends(SampleNow, _super);
    function SampleNow(b) {
        _super.call(this);
        this.b = b;
    }
    SampleNow.prototype.run = function () {
        return behavior_1.at(this.b);
    };
    return SampleNow;
}(Now));
function sample(b) {
    return new SampleNow(b);
}
exports.sample = sample;
var PerformIOStream = (function (_super) {
    __extends(PerformIOStream, _super);
    function PerformIOStream(s) {
        _super.call(this);
        s.addListener(this);
    }
    PerformIOStream.prototype.push = function (io) {
        var _this = this;
        io_1.runIO(io).then(function (a) { return _this.child.push(a); });
    };
    return PerformIOStream;
}(stream_1.Stream));
var PerformStreamNow = (function (_super) {
    __extends(PerformStreamNow, _super);
    function PerformStreamNow(s) {
        _super.call(this);
        this.s = s;
    }
    PerformStreamNow.prototype.run = function () {
        return new PerformIOStream(this.s);
    };
    return PerformStreamNow;
}(Now));
function performStream(s) {
    return new PerformStreamNow(s);
}
exports.performStream = performStream;
var PerformIOStreamLatest = (function (_super) {
    __extends(PerformIOStreamLatest, _super);
    function PerformIOStreamLatest(s) {
        _super.call(this);
        this.id = 0;
        this.latest = 0;
        s.addListener(this);
    }
    PerformIOStreamLatest.prototype.push = function (io) {
        var _this = this;
        var id = this.id++;
        io_1.runIO(io).then(function (a) {
            if (id > _this.latest) {
                _this.latest = id;
                _this.child.push(a);
            }
        });
    };
    return PerformIOStreamLatest;
}(stream_1.Stream));
var PerformStreamNowLatest = (function (_super) {
    __extends(PerformStreamNowLatest, _super);
    function PerformStreamNowLatest(s) {
        _super.call(this);
        this.s = s;
    }
    PerformStreamNowLatest.prototype.run = function () {
        return new PerformIOStreamLatest(this.s);
    };
    return PerformStreamNowLatest;
}(Now));
function performStreamLatest(s) {
    return new PerformStreamNowLatest(s);
}
exports.performStreamLatest = performStreamLatest;
var PerformIOStreamOrdered = (function (_super) {
    __extends(PerformIOStreamOrdered, _super);
    function PerformIOStreamOrdered(s) {
        _super.call(this);
        this.nextId = 0;
        this.next = 0;
        this.buffer = [];
        s.addListener(this);
    }
    PerformIOStreamOrdered.prototype.push = function (io) {
        var _this = this;
        var id = this.nextId++;
        io_1.runIO(io).then(function (a) {
            if (id === _this.next) {
                _this.buffer[0] = a;
                _this.pushFromBuffer();
            }
            else {
                _this.buffer[id - _this.next] = a;
            }
        });
    };
    PerformIOStreamOrdered.prototype.pushFromBuffer = function () {
        while (this.buffer[0] !== undefined) {
            var element = this.buffer.shift();
            this.child.push(element);
            this.next++;
        }
    };
    return PerformIOStreamOrdered;
}(stream_1.Stream));
var PerformStreamNowOrdered = (function (_super) {
    __extends(PerformStreamNowOrdered, _super);
    function PerformStreamNowOrdered(s) {
        _super.call(this);
        this.s = s;
    }
    PerformStreamNowOrdered.prototype.run = function () {
        return new PerformIOStreamOrdered(this.s);
    };
    return PerformStreamNowOrdered;
}(Now));
function performStreamOrdered(s) {
    return new PerformStreamNowOrdered(s);
}
exports.performStreamOrdered = performStreamOrdered;
function run(now) {
    return now.run();
}
var PlanNow = (function (_super) {
    __extends(PlanNow, _super);
    function PlanNow(future) {
        _super.call(this);
        this.future = future;
    }
    PlanNow.prototype.run = function () {
        return this.future.map(run);
    };
    return PlanNow;
}(Now));
function plan(future) {
    return new PlanNow(future);
}
exports.plan = plan;
function runNow(now) {
    return new Promise(function (resolve, reject) {
        now.run().subscribe(resolve);
    });
}
exports.runNow = runNow;
//# sourceMappingURL=now.js.map