"use strict";
var behavior_1 = require("./behavior");
var stream_1 = require("./stream");
var Placeholder = (function () {
    function Placeholder(methodName) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        this.methodName = methodName;
        this.children = [];
        this.args = args;
    }
    Placeholder.prototype.replaceWith = function (parent) {
        this.source = this.methodName === undefined ? parent : parent[this.methodName].apply(parent, this.args);
        for (var _i = 0, _a = this.children; _i < _a.length; _i++) {
            var c = _a[_i];
            c.replaceWith(this.source);
        }
        this.children = undefined;
    };
    return Placeholder;
}());
exports.Placeholder = Placeholder;
function definePlaceholderMethod(methodName) {
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i - 0] = arguments[_i];
        }
        if (this.source !== undefined) {
            return (_a = this.source)[methodName].apply(_a, args);
        }
        else {
            var p = new (Placeholder.bind.apply(Placeholder, [void 0, methodName].concat(args)))();
            this.children.push(p);
            return p;
        }
        var _a;
    };
}
function defineBehaviorMethod(methodName) {
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i - 0] = arguments[_i];
        }
        if (this.source !== undefined) {
            return (_a = this.source)[methodName].apply(_a, args);
        }
        else {
            var p = behavior_1.placeholder();
            this.children.push(p);
            return p[methodName].apply(p, args);
        }
        var _a;
    };
}
function defineStreamMethod(methodName) {
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i - 0] = arguments[_i];
        }
        if (this.source !== undefined) {
            return (_a = this.source)[methodName].apply(_a, args);
        }
        else {
            var p = stream_1.placeholderStream();
            this.children.push(p);
            return p[methodName].apply(p, args);
        }
        var _a;
    };
}
var commonMethods = [
    "map", "mapTo", "subscribe", "addListener"
];
var streamMethods = [
    "combine", "filter", "filterApply", "scanS", "delay", "throttle", "debounce"
];
var behaviorMethods = [
    "ap", "lift", "chain", "push", "pull", "beginPulling", "endPulling",
    "observe", "at", "flatten"
];
for (var _i = 0, commonMethods_1 = commonMethods; _i < commonMethods_1.length; _i++) {
    var name = commonMethods_1[_i];
    Placeholder.prototype[name] = definePlaceholderMethod(name);
}
for (var _a = 0, streamMethods_1 = streamMethods; _a < streamMethods_1.length; _a++) {
    var name = streamMethods_1[_a];
    Placeholder.prototype[name] = defineStreamMethod(name);
}
for (var _b = 0, behaviorMethods_1 = behaviorMethods; _b < behaviorMethods_1.length; _b++) {
    var name = behaviorMethods_1[_b];
    Placeholder.prototype[name] = defineBehaviorMethod(name);
}
function placeholder() {
    return new Placeholder();
}
exports.placeholder = placeholder;
//# sourceMappingURL=placeholder.js.map